use crate::{
    token::{Token, TokenType::Eof},
    tokenizer::{
        comment::Comment, identifier::Identifier, interface::Tokenizer, lookahead::Lookahead,
        newline::Newline, number::Number, single_char::SingleChar, string::String,
        whitespace::Whitespace,
    },
};
use lazy_static::lazy_static;
use tracing::{debug, error};

lazy_static! {
    static ref TOKENIZERS: Vec<Box<dyn Tokenizer>> = vec![
        Box::new(SingleChar),
        Box::new(Whitespace),
        Box::new(Newline),
        Box::new(Lookahead),
        Box::new(Comment),
        Box::new(String),
        Box::new(Number),
        Box::new(Identifier),
    ];
}

pub fn tokenize(source: &str) -> Vec<Token> {
    let mut tokens: Vec<Token> = Vec::new();
    let mut source_chars = source.char_indices().peekable();

    let mut line = 1;
    while let Some(c) = source_chars.next() {
        let mut tokenizer_idx = 0;
        let mut tokenizer_result = None;
        while tokenizer_idx < TOKENIZERS.len() && tokenizer_result.is_none() {
            tokenizer_result = TOKENIZERS[tokenizer_idx].run(c, &mut source_chars, source, line);
            tokenizer_idx += 1;
        }

        match tokenizer_result {
            Some((line_advance, token)) => {
                line += line_advance;

                if let Some(token) = token {
                    tokens.push(token);
                }
            }
            None => {
                error!("unexpected character: {}", c.1)
            }
        }
    }
    tokens.push(Token {
        token_type: Eof,
        lexeme: "".to_string(),
        literal: None,
        line,
    });
    debug!("{:?}", tokens);

    tokens
}
